#!/usr/bin/env python3
"""Calculate different energy contributions."""
import dataclasses

import numpy as np
from scipy.linalg import inv, norm
from scipy.special import erfc

from .dft import get_n_single, H, solve_poisson
from .extras import dispersion
from .gga import get_grad_field, get_tau
from .logger import log
from .tools import electronic_entropy
from .utils import handle_k_reducable
from .xc import get_exc


@dataclasses.dataclass
class Energy:
    """Energy class to save energy contributions in one place."""
    Ekin: float = 0     #: Kinetic energy.
    Ecoul: float = 0    #: Coulomb energy.
    Exc: float = 0      #: Exchange-correlation energy.
    Eloc: float = 0     #: Local energy.
    Enonloc: float = 0  #: Non-local energy.
    Eewald: float = 0   #: Ewald energy.
    Esic: float = 0     #: Self-interaction correction energy.
    Edisp: float = 0    #: Dispersion correction energy.
    Ets: float = 0      #: Fillings entropic energy.

    @property
    def Etot(self):
        """Total energy is the sum of all energy contributions."""
        return sum(getattr(self, ie.name) for ie in dataclasses.fields(self))

    def __repr__(self):
        """Print the energies stored in the Energy object."""
        out = ''
        for ie in dataclasses.fields(self):
            energy = getattr(self, ie.name)
            if energy != 0:
                out += f'{ie.name:<8}: {energy:+.9f} Eh\n'
        return f'{out}{"-" * 25}\nEtot    : {self.Etot:+.9f} Eh'


def get_E(scf):
    """Calculate energy contributions and update energies needed in one SCF step.

    Args:
        scf: SCF object.

    Returns:
        float: Total energy.
    """
    scf.energies.Ekin = get_Ekin(scf.atoms, scf.Y)
    scf.energies.Ecoul = get_Ecoul(scf.atoms, scf.n, scf.phi)
    scf.energies.Exc = get_Exc(scf, scf.n, scf.exc, Nspin=scf.atoms.occ.Nspin)
    scf.energies.Eloc = get_Eloc(scf, scf.n)
    scf.energies.Enonloc = get_Enonloc(scf, scf.Y)
    return scf.energies.Etot


@handle_k_reducable
def get_Ekin(atoms, Y, ik):
    """Calculate the kinetic energy.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.
        ik (int): k-point index.

    Returns:
        float: Kinetic energy in Hartree.
    """
    # Ekin = -0.5 Tr(F Wdag L(W))
    Ekin = 0
    for spin in range(atoms.occ.Nspin):
        Ekin += -0.5 * atoms.kpts.wk[ik] * np.trace(atoms.occ.F[ik][spin] @ Y[spin].conj().T @
                                                    atoms.L(Y, ik)[spin])
    return np.real(Ekin)


def get_Ecoul(atoms, n, phi=None):
    """Calculate the Coulomb energy.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        n (ndarray): Real-space electronic density.

    Keyword Args:
        phi (ndarray | None): Hartree field.

    Returns:
        float: Coulomb energy in Hartree.
    """
    if phi is None:
        phi = solve_poisson(atoms, n)
    # Ecoul = 0.5 (J(n))dag O(phi)
    return np.real(0.5 * n @ atoms.Jdag(atoms.O(phi)))


def get_Exc(scf, n, exc=None, n_spin=None, dn_spin=None, tau=None, Nspin=2):
    """Calculate the exchange-correlation energy.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        n (ndarray): Real-space electronic density.

    Keyword Args:
        exc (ndarray | None): Exchange-correlation energy density.
        n_spin (ndarray | None): Real-space electronic densities per spin channel.
        dn_spin (ndarray | None): Real-space gradient of densities per spin channel.
        tau (ndarray | None): Real-space kinetic energy densities per spin channel.
        Nspin (int): Number of spin states.

    Returns:
        float: Exchange-correlation energy in Hartree.
    """
    atoms = scf.atoms
    if exc is None:
        exc = get_exc(scf.xc, n_spin, Nspin, dn_spin, tau)
    # Exc = (J(n))dag O(J(exc))
    return np.real(n @ atoms.Jdag(atoms.O(atoms.J(exc))))


def get_Eloc(scf, n):
    """Calculate the local energy contribution.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        n (ndarray): Real-space electronic density.

    Returns:
        float: Local energy contribution in Hartree.
    """
    return np.real(scf.Vloc.conj().T @ n)


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/calc_energies.jl
@handle_k_reducable
def get_Enonloc(scf, Y, ik):
    """Calculate the non-local GTH energy contribution.

    Reference: Phys. Rev. B 54, 1703.

    Args:
        scf: SCF object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.
        ik (int): k-point index.

    Returns:
        float: Non-local GTH energy contribution in Hartree.
    """
    atoms = scf.atoms

    Enonloc = 0
    if scf.pot == 'gth' and scf.gth.NbetaNL > 0:  # Only calculate the non-local part if necessary
        for spin in range(atoms.occ.Nspin):
            betaNL_psi = (Y[spin].conj().T @ scf.gth.betaNL[ik]).conj()

            enl = np.zeros(Y.shape[-1], dtype=complex)
            for ia in range(atoms.Natoms):
                psp = scf.gth[atoms.atom[ia]]
                for l in range(psp['lmax']):
                    for m in range(-l, l + 1):
                        for iprj in range(psp['Nproj_l'][l]):
                            ibeta = scf.gth.prj2beta[iprj, ia, l, m + psp['lmax'] - 1] - 1
                            for jprj in range(psp['Nproj_l'][l]):
                                jbeta = scf.gth.prj2beta[jprj, ia, l, m + psp['lmax'] - 1] - 1
                                hij = psp['h'][l, iprj, jprj]
                                enl += hij * betaNL_psi[:, ibeta].conj() * betaNL_psi[:, jbeta]
            Enonloc += np.sum(atoms.occ.f[ik, spin] * atoms.kpts.wk[ik] * enl)
    # We have to multiply with the cell volume, because of different orthogonalization methods
    return np.real(Enonloc * atoms.Omega)


def get_Eewald(atoms, gcut=2, gamma=1e-8):
    """Calculate the Ewald energy.

    Reference: J. Chem. Theory Comput. 10, 381.

    Args:
        atoms: Atoms object.

    Keyword Args:
        gcut (float): G-vector cut-off.
        gamma (float): Error tolerance.

    Returns:
        float: Ewald energy in Hartree.
    """
    # For a plane wave code we have multiple contributions for the Ewald energy
    # Namely, a sum from contributions from real-space, reciprocal space,
    # the self energy, (the dipole term [neglected]), and an additional electroneutrality term
    def get_index_vectors(s):
        """Create index vectors of s periodic images."""
        m1 = np.arange(-s[0], s[0] + 1)
        m2 = np.arange(-s[1], s[1] + 1)
        m3 = np.arange(-s[2], s[2] + 1)
        M = np.transpose(np.meshgrid(m1, m2, m3)).reshape(-1, 3)
        # Delete the [0, 0, 0] element, to prevent checking for it in every loop
        return M[~(M == 0).all(axis=1)]

    # Calculate the Ewald splitting parameter nu
    gexp = -np.log(gamma)
    nu = 0.5 * np.sqrt(gcut**2 / gexp)

    # Start by calculating the self-energy
    Eewald = -nu / np.sqrt(np.pi) * np.sum(atoms.Z**2)
    # Add the electroneutrality term
    Eewald += -np.pi * np.sum(atoms.Z)**2 / (2 * nu**2 * atoms.Omega)

    # Calculate the real-space contribution
    # Calculate the amount of images that have to be considered per axis
    Rm = norm(atoms.a, axis=1)
    tmax = np.sqrt(0.5 * gexp) / nu
    s = np.rint(tmax / Rm + 1.5)
    # Collect all box index vectors in a matrix
    M = get_index_vectors(s)
    # Calculate the translation vectors
    T = M @ atoms.a

    # Calculate the reciprocal space contribution
    # Calculate the amount of reciprocal images that have to be considered per axis
    g = 2 * np.pi * inv(atoms.a.T)
    gm = norm(g, axis=1)
    s = np.rint(gcut / gm + 1.5)
    # Collect all box index vectors in a matrix
    M = get_index_vectors(s)
    # Calculate the reciprocal translation vectors and precompute the prefactor
    G = M @ g
    G2 = norm(G, axis=1)**2
    prefactor = 2 * np.pi / atoms.Omega * np.exp(-0.25 * G2 / nu**2) / G2

    for ia in range(atoms.Natoms):
        for ja in range(atoms.Natoms):
            dpos = atoms.pos[ia] - atoms.pos[ja]
            ZiZj = atoms.Z[ia] * atoms.Z[ja]

            # Add the real-space contribution
            rmag = np.sqrt(norm(dpos - T, axis=1)**2)
            Eewald += 0.5 * ZiZj * np.sum(erfc(rmag * nu) / rmag)
            # The T=[0, 0, 0] element is omitted in M but needed if ia!=ja, so add it manually
            if ia != ja:
                rmag = np.sqrt(norm(dpos)**2)
                Eewald += 0.5 * ZiZj * erfc(rmag * nu) / rmag

            # Add the reciprocal space contribution
            Gpos = np.sum(G * dpos, axis=1)
            Eewald += ZiZj * np.sum(prefactor * np.cos(Gpos))
    return Eewald


def get_Esic(scf, Y, n_single=None):
    """Calculate the Perdew-Zunger self-interaction energy.

    Reference: Phys. Rev. B 23, 5048.

    Args:
        scf: SCF object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Keyword Args:
        n_single (ndarray | None): Single-electron densities.

    Returns:
        float: PZ self-interaction energy.
    """
    atoms = scf.atoms
    # E_PZ-SIC = \sum_i Ecoul[n_i] + Exc[n_i, 0]
    if n_single is None:
        n_single = get_n_single(atoms, Y)

    Esic = 0
    for i in range(atoms.occ.Nstate):
        for spin in range(atoms.occ.Nspin):
            if np.sum(atoms.occ.f[:, spin, i] * atoms.kpts.wk) > 0:
                # Create normalized single-particle densities in the form of electronic densities
                # per spin channel, since spin-polarized functionals expect this form
                ni = np.zeros((2, atoms.Ns))
                # Normalize single-particle densities to 1
                ni[0] = n_single[spin, :, i] / np.sum(atoms.occ.f[:, spin, i] * atoms.kpts.wk)

                # Get the gradient of the single-particle density
                if 'gga' in scf.xc_type:
                    dni = np.zeros((2, atoms.Ns, 3))
                    dni[0] = get_grad_field(atoms, ni)[0]
                else:
                    dni = None

                # Get the kinetic energy density of the corresponding orbital
                if scf.xc_type == 'meta-gga':
                    # Use only one orbital for the calculation
                    Ytmp = []
                    for ik in range(atoms.kpts.Nk):
                        Ytmp.append(np.zeros_like(Y[ik]))
                        Ytmp[ik][0, :, 0] = Y[ik][spin, :, i]
                    taui = np.zeros_like(ni)
                    # We also have to normalize to one again
                    taui[0] = get_tau(atoms, Ytmp)[0] / np.sum(atoms.occ.f[:, spin, i] *
                                                               atoms.kpts.wk)
                else:
                    taui = None

                coul = get_Ecoul(atoms, ni[0])
                # The exchange part for a SIC correction has to be spin-polarized
                xc = get_Exc(scf, ni[0], n_spin=ni, dn_spin=dni, tau=taui, Nspin=2)
                # SIC energy is scaled by the occupation number
                Esic += (coul + xc) * np.sum(atoms.occ.f[:, spin, i] * atoms.kpts.wk)
    scf.energies.Esic = Esic
    return Esic


def get_Edisp(scf, version='d3bj', atm=True, xc=None):  # noqa: D103
    try:
        return dispersion.get_Edisp(scf, version, atm, xc)
    except ImportError:
        log.warning('You have to install the dispersion extra to use this function.')
        return 0


get_Edisp.__doc__ = dispersion.get_Edisp.__doc__


def get_Eband(scf, Y, **kwargs):
    """Calculate the band energy for occupied or unoccupied bands.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Keyword Args:
        **kwargs: See :func:`eminus.dft.H`.

    Returns:
        float: Band energy in Hartree.
    """
    atoms = scf.atoms
    # Eband = Tr(Wdag H(W))
    Eband = 0
    for ik in range(atoms.kpts.Nk):
        for spin in range(atoms.occ.Nspin):
            Eband += atoms.kpts.wk[ik] * np.trace(Y[ik][spin].conj().T @ H(scf, ik, spin, Y,
                                                                           **kwargs))
    return np.real(Eband)


def get_Ets(scf, epsilon, Efermi):
    """Calculate the fillings entropic energy.

    Reference: https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/occupations.jl

    Args:
        scf: SCF object.
        epsilon (ndarray): Eigenenergies.
        Efermi (float): Fermi energy.

    Returns:
        float: Entropic energy in Hartree.
    """
    occ = scf.atoms.occ

    wk = occ.wk
    if occ.Nspin == 1:
        wk = 2 * wk

    Ets = 0
    for ik in range(len(wk)):
        for spin in range(occ.Nspin):
            for istate in range(occ.Nstate):
                Ets -= wk[ik] * occ.smearing * electronic_entropy(epsilon[ik, spin, istate], Efermi,
                                                                  occ.smearing)
    scf.energies.Ets = Ets
    return Ets



#!/usr/bin/env python3
"""Cell wrapper function."""
import numpy as np

from .atoms import Atoms
from .data import LATTICE_VECTORS
from .utils import molecule2list

#: Crystal structures with their respective lattice and basis.
STRUCTURES = {
    'sc': {
        'lattice': 'sc',
        'basis': [[0, 0, 0]]
    },
    'fcc': {
        'lattice': 'fcc',
        'basis': [[0, 0, 0]]
    },
    'bcc': {
        'lattice': 'bcc',
        'basis': [[0, 0, 0]]
    },
    'tetragonal': {
        'lattice': 'sc',
        'basis': [[0, 0, 0]]
    },
    'orthorhombic': {
        'lattice': 'sc',
        'basis': [[0, 0, 0]]
    },
    'hexagonal': {
        'lattice': 'hexagonal',
        'basis': [[0, 0, 0]]
    },
    'diamond': {
        'lattice': 'fcc',
        'basis': [[0, 0, 0], [1 / 4, 1 / 4, 1 / 4]]
    },
    'zincblende': {
        'lattice': 'fcc',
        'basis': [[0, 0, 0], [1 / 4, 1 / 4, 1 / 4]]
    },
    'rocksalt': {
        'lattice': 'fcc',
        'basis': [[0, 0, 0], [1 / 2, 0, 0]]
    },
    'cesiumchloride': {
        'lattice': 'sc',
        'basis': [[0, 0, 0], [1 / 2, 1 / 2, 1 / 2]]
    },
    'fluorite': {
        'lattice': 'fcc',
        'basis': [[0, 0, 0], [1 / 4, 1 / 4, 1 / 4], [3 / 4, 3 / 4, 3 / 4]]
    }
}


def Cell(atom, lattice, ecut, a, basis=None, bands=None, kmesh=1, smearing=0, verbose=None):
    """Wrapper to create Atoms classes for crystal systems.

    Args:
        atom (str | list | tuple): Atom symbols.
        lattice (str | list | tuple | ndarray): Lattice system.
        ecut (float):  Cut-off energy.
        a (float | list | tuple | ndarray | None): Cell size.

    Keyword Args:
        basis (list | tuple | ndarray | None): Lattice basis.
        bands (int | None): Number of bands (has to be larger or equal than the occupied states).
        kmesh (int | list | tuple | ndarray): k-point mesh.
        smearing (float): Smearing width.
        verbose (int | str | None): Level of output.

    Returns:
        Atoms object.
    """
    # Get the lattice vectors from a string or use them directly
    if isinstance(lattice, str):
        lattice = lattice.lower()
        if basis is None:
            basis = STRUCTURES[lattice]['basis']
        lattice = STRUCTURES[lattice]['lattice']
        lattice_vectors = np.asarray(LATTICE_VECTORS[lattice])
    else:
        if basis is None:
            basis = [[0, 0, 0]]
        lattice_vectors = np.asarray(lattice)

    # Only scale the lattice vectors with if a is given
    if a is not None:
        lattice_vectors = a * lattice_vectors
        basis = a * np.asarray(basis)

    # Account for different atom and basis sizes
    if isinstance(atom, str):
        atom_list = molecule2list(atom)
    else:
        atom_list = atom
    if len(atom_list) != len(basis):
        atom = [atom] * len(basis)

    # Build the atoms object
    atoms = Atoms(atom, basis, ecut=ecut, a=lattice_vectors, verbose=verbose)
    # Handle k-points and states
    atoms.kpts.kmesh = kmesh
    if isinstance(lattice, str):
        atoms.kpts.lattice = lattice
    atoms.occ.smearing = smearing
    if bands is not None:
        atoms.occ.bands = bands
    return atoms



#!/usr/bin/env python3
"""Extra functions that need additional dependencies to work.

To also install all additional dependencies, use::

    pip install eminus[all]

Alternatively, you can only install selected extras using the respective name:

* :mod:`~eminus.extras.dispersion`
* :mod:`~eminus.extras.fods`
* :mod:`~eminus.extras.libxc`
* :mod:`~eminus.extras.symmetry`
* :mod:`~eminus.extras.torch`
* :mod:`~eminus.extras.viewer`
"""
from . import torch
from .dispersion import get_Edisp
from .fods import get_fods, remove_core_fods, split_fods
from .libxc import libxc_functional
from .symmetry import symmetrize
from .viewer import (
    executed_in_notebook,
    plot_bandstructure,
    view,
    view_atoms,
    view_contour,
    view_file,
    view_kpts,
)

__all__ = ['executed_in_notebook', 'get_Edisp', 'get_fods', 'libxc_functional',
           'plot_bandstructure', 'remove_core_fods', 'split_fods', 'symmetrize', 'torch', 'view',
           'view_atoms', 'view_contour', 'view_file', 'view_kpts']



#!/usr/bin/env python3
"""Viewer functions for Jupyter notebooks."""
import io
import pathlib
import uuid

import numpy as np

from ..atoms import Atoms
from ..data import COVALENT_RADII, CPK_COLORS, SPECIAL_POINTS
from ..dft import get_epsilon, get_epsilon_unocc
from ..io import create_pdb_str, read_cube, read_traj, read_xyz
from ..kpoints import get_brillouin_zone, kpoint_convert, KPoints, kpoints2axis
from ..logger import log
from ..occupations import find_Efermi
from ..tools import get_Efermi, get_isovalue
from ..units import ha2ev
from ..utils import get_lattice
from .fods import split_fods


def view(*args, **kwargs):
    """Unified display function."""
    if isinstance(args[0], (str, list, tuple)):
        return view_file(*args, **kwargs)
    if isinstance(args[0], KPoints):
        return view_kpts(*args, **kwargs)
    return view_atoms(*args, **kwargs)


def view_atoms(obj, extra=None, plot_n=False, percent=85, surfaces=20, size=(600, 600)):
    """Display atoms with optional FODs or grid points, or even densities.

    Reference: https://plotly.com/python

    Args:
        obj: Atoms or SCF object.

    Keyword Args:
        extra (ndarray | list | None): Extra coordinates or FODs to display.
        plot_n (bool): Whether to plot the electronic density (only for executed scf objects).
        percent (float): Amount of density that should be contained.
        surfaces (int): Number of surfaces to display in density plots (reduce for performance).
        size (tuple): Widget size.

    Returns:
        Figure: Viewable object.
    """
    try:
        import plotly.graph_objects as go
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise
    atoms = obj._atoms

    fig = go.Figure()
    # Calculate vertices of the Brillouin zone and plot them
    lattice = get_lattice(atoms.a)
    for xx in lattice:
        bz_data = go.Scatter3d(x=xx[:, 0], y=xx[:, 1], z=xx[:, 2],
                               name='Unit cell',
                               showlegend=False,
                               marker={'size': 0.1, 'color': 'black'})
        fig.add_trace(bz_data)
    # Add species one by one to be able to have them named and be selectable in the legend
    # Note: The size scaling is mostly arbitrary and has no meaning
    for ia in sorted(set(atoms.atom)):
        mask = np.where(np.asarray(atoms.atom) == ia)[0]
        atom_data = go.Scatter3d(x=atoms.pos[mask, 0], y=atoms.pos[mask, 1], z=atoms.pos[mask, 2],
                                 name=ia,
                                 mode='markers',
                                 marker={'size': 2 * np.pi * np.sqrt(COVALENT_RADII[ia]),
                                         'color': CPK_COLORS[ia],
                                         'line': {'color': 'black', 'width': 2}})
        fig.add_trace(atom_data)
    if extra is not None:
        # If a list has been provided with a length of two it has to be FODs
        if isinstance(extra, list):
            if len(extra[0]) != 0:
                extra_data = go.Scatter3d(x=extra[0][:, 0], y=extra[0][:, 1], z=extra[0][:, 2],
                                          name='up-FOD',
                                          mode='markers',
                                          marker={'size': np.pi, 'color': 'red'})
                fig.add_trace(extra_data)
            if len(extra) > 1 and len(extra[1]) != 0:
                extra_data = go.Scatter3d(x=extra[1][:, 0], y=extra[1][:, 1], z=extra[1][:, 2],
                                          name='down-FOD',
                                          mode='markers',
                                          marker={'size': np.pi, 'color': 'green'})
                fig.add_trace(extra_data)
        # Treat extra as normal coordinates otherwise
        else:
            extra_data = go.Scatter3d(x=extra[:, 0], y=extra[:, 1], z=extra[:, 2],
                                      name='Coordinates',
                                      mode='markers',
                                      marker={'size': 1, 'color': 'red'})
            fig.add_trace(extra_data)

    # A density can be plotted for an SCF object
    if isinstance(plot_n, np.ndarray) or plot_n:
        # Plot a given array or use the density from an SCF object
        if isinstance(plot_n, np.ndarray):
            density = plot_n
        else:
            density = obj.n
        den_data = go.Volume(x=atoms.r[:, 0], y=atoms.r[:, 1], z=atoms.r[:, 2], value=density,
                             name='Density',
                             colorbar_title=f'Density ({percent}%)',
                             colorscale='Inferno',
                             isomin=get_isovalue(density, percent=percent),
                             isomax=np.max(density),
                             surface_count=surfaces,
                             opacity=0.1,
                             showlegend=True)
        fig.add_trace(den_data)
        # Move colorbar to the left
        fig.data[-1].colorbar.x = -0.15

    # Theming
    scene = {'xaxis': {'title': 'x [a<sub>0</sub>]'},
             'yaxis': {'title': 'y [a<sub>0</sub>]'},
             'zaxis': {'title': 'z [a<sub>0</sub>]'},
             'aspectmode': 'cube'}
    # If the unit cell is diagonal and we scale the plot, otherwise let plotly decide
    if (np.diag(np.diag(atoms.a)) == atoms.a).all():
        scene['xaxis']['range'] = [0, atoms.a[0, 0]]
        scene['yaxis']['range'] = [0, atoms.a[1, 1]]
        scene['zaxis']['range'] = [0, atoms.a[2, 2]]
    fig.update_layout(
        scene=scene,
        width=size[0],
        height=size[1],
        legend={'itemsizing': 'constant', 'title': 'Selection'},
        hoverlabel_bgcolor='black',
        template='none')
    if executed_in_notebook():
        return fig
    return fig.show()


def view_contour(obj, field, axis=2, value=0.5, lines=10, limits=(-1, 1), zoom=1, linewidth=1,
                 size=(600, 600)):
    """Display contour lines of field data like electronic densities.

    Reference: https://plotly.com/python

    Args:
        obj: Atoms or SCF object.
        field: Real-space field data.

    Keyword Args:
        axis (int): Axis to slice through.
        value: (float): Slice value scaled by the axis size.
        lines: Number of contour lines.
        limits (tuple): Minimal and maximal truncation values for the field data.
        zoom (float): Initial zoom.
        linewidth (float): Contour line width.
        size (tuple): Widget size.

    Returns:
        Figure: Viewable object.
    """
    try:
        import plotly.graph_objects as go
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise
    atoms = obj._atoms

    # Get the axes indices
    axes = [(axis + 1) % 3, (axis + 2) % 3, axis]
    # Create an index mask to obtain a clean slice through the cell
    # Since the expression is a bit involved it can be understood with the following pseudo-code:
    # mask = |axis_values - slice_value| < axis_value_dist
    mask = np.abs(atoms.r[:, axes[2]] - value * atoms.a[axes[2], axes[2]]) < \
        (atoms.a[axes[2], axes[2]] / atoms.s[axes[2]])
    # Create a copy of the field data to not overwrite the input
    field = np.copy(field)
    # Remove large and small values (similar to VESTA)
    field[field < limits[0]] = limits[0]
    field[field > limits[1]] = limits[1]

    # Create the contour lines
    fig = go.Figure()
    contours = go.Contour(x=atoms.r[:, axes[0]][mask], y=atoms.r[:, axes[1]][mask], z=field[mask],
                          contours_coloring='none',
                          ncontours=lines,
                          line_width=linewidth)
    fig.add_trace(contours)

    # Theming
    fig.update_layout(
        width=size[0],
        height=size[1],
        margin={'b': 0, 'l': 0, 'r': 0, 't': 0},
        xaxis={'range': [(1 - 1 / zoom) * atoms.a[axes[0], axes[0]],
                         1 / zoom * atoms.a[axes[0], axes[0]]],
               'visible': False},
        yaxis={'range': [(1 - 1 / zoom) * atoms.a[axes[1], axes[1]],
                         1 / zoom * atoms.a[axes[1], axes[1]]],
               'visible': False},
        template='none')
    if executed_in_notebook():
        return fig
    return fig.show()


def view_file(filename, isovalue=0.01, gui=False, elec_symbols=('X', 'He'),
              size=(400, 400), **kwargs):
    """Display molecules and orbitals.

    Reference: Bioinformatics 34, 1241.

    Args:
        filename (str | list | tuple): Input filename(s). This can be either CUBE or XYZ files.

    Keyword Args:
        isovalue (float): Isovalue for sizing orbitals.
        gui (bool): Turn on the NGLView GUI.
        elec_symbols (list): Identifier for up and down FODs.
        size (tuple): Widget size.

    Keyword Args:
        **kwargs: NGLWidget keyword arguments.

    Returns:
        NGLWidget: Viewable object.
    """
    try:
        from nglview import NGLWidget, write_html
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise

    # If multiple files are given, try to open them with an interact drop-down menu
    # If all files are XYZ files they will be displayed as a trajectory instead
    if isinstance(filename, (list, tuple)) and not np.all([f.endswith('.xyz')for f in filename]):
        if executed_in_notebook():
            from ipywidgets import interact
            interact(lambda filename: view_file(filename, isovalue, gui, elec_symbols,
                                                **kwargs), filename=filename)
            return None
        # If we are not in a notebook open the files one by one
        for f in filename:
            view_file(f, isovalue, gui, elec_symbols, **kwargs)
            return None

    view = NGLWidget(**kwargs)
    view._set_size(f'{size[0]}px', f'{size[1]}px')
    # Set the gui to the view
    if gui:
        view.gui_style = 'ngl'

    # Handle TRAJ files (or multiple XYZ files)
    if isinstance(filename, (list, tuple)) or filename.endswith(('.trj', '.traj')):
        view = _traj_view(view, filename)
    # Handle XYZ files
    elif filename.endswith('.xyz'):
        view = _xyz_view(view, filename, elec_symbols)
    # Handle CUBE files
    elif filename.endswith(('.cub', '.cube')):
        view = _cube_view(view, filename, isovalue, elec_symbols)
    # Handle other files (mainly for PDBs)
    else:
        view = _generic_view(view, filename)

    # Check if the function has been executed in a notebook
    # If yes, just return the view
    if executed_in_notebook():
        return view
    # Otherwise the viewer would display nothing
    # But if plotly is installed on can write the view to a StringIO object and display it with the
    # smart open_html_in_browser function from plotly that automatically opens a new browser tab
    try:
        from plotly.io._base_renderers import open_html_in_browser
        # Use StringIO object instead of a temporary file
        with io.StringIO() as html:
            write_html(html, view)
            open_html_in_browser(html.getvalue())
    except ImportError:
        log.error('This function only works in notebooks or with Plotly installed.')
    return None


def executed_in_notebook():
    """Check if the code runs in a notebook.

    Reference: https://stackoverflow.com/a/22424821

    Returns:
        bool: Whether in a notebook or not.
    """
    try:
        from IPython import get_ipython
        if 'IPKernelApp' not in get_ipython().config:
            return False
    except (AttributeError, ImportError):
        return False
    return True


def _generic_view(view, filename):
    """Modify the view for a generic (probably PDB) file.

    Args:
        view (NGLWidget): Viewable object.
        filename (str): Input filename.

    Returns:
        NGLWidget: Viewable object.
    """
    # If no XYZ or CUBE file is used try a more generic file viewer
    ext = pathlib.Path(filename).suffix.replace('.', '')
    # It seems that only PDB works with this method
    if ext != 'pdb':
        log.warning('Only XYZ, CUBE, and PDB files are support, but others might work.')
    with open(filename, encoding='utf-8') as fh:
        view.add_component(fh, ext=ext)
    view[0].clear()
    view[0].add_ball_and_stick()
    view.center()
    return view


def _cube_view(view, filename, isovalue, elec_symbols):
    """Modify the view for a given CUBE file.

    Args:
        view (NGLWidget): Viewable object.
        filename (str | list | tuple): Input filename(s).
        isovalue (float): Isovalue for sizing orbitals.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        NGLWidget: Viewable object.
    """
    from nglview import TextStructure
    # Atoms and cell
    atom, pos, _, a, _, _ = read_cube(filename)
    atom, pos, pos_fod = split_fods(atom, pos, elec_symbols)
    view.add_component(TextStructure(create_pdb_str(atom, pos, a)))
    view[0].clear()
    view[0].add_ball_and_stick()
    view.add_unitcell()
    view.center()
    # Spin up
    view.add_component(filename)
    view[1].clear()
    view[1].add_surface(negateIsolevel=False,
                        isolevelType='value',
                        isolevel=isovalue,
                        color='lightgreen',
                        opacity=0.75,
                        side='front',
                        depthWrite=False)
    # Spin down
    view.add_component(filename)
    view[2].clear()
    view[2].add_surface(negateIsolevel=True,
                        isolevelType='value',
                        isolevel=isovalue,
                        color='red',
                        opacity=0.75,
                        side='front',
                        depthWrite=False)
    # Spin up FODs
    if len(pos_fod[0]) > 0:
        view.add_component(TextStructure(create_pdb_str([elec_symbols[0]] * len(pos_fod[0]),
                           pos_fod[0])))
        view[3].clear()
        view[3].add_ball_and_stick(f'_{elec_symbols[0]}', color='red', radius=0.1)
    # Spin down FODs
    if len(pos_fod[1]) > 0:
        view.add_component(TextStructure(create_pdb_str([elec_symbols[1]] * len(pos_fod[1]),
                           pos_fod[1])))
        view[4].clear()
        view[4].add_ball_and_stick(f'_{elec_symbols[1]}', color='green', radius=0.1)
    return view


def _xyz_view(view, filename, elec_symbols):
    """Modify the view for a given XYZ file.

    Args:
        view (NGLWidget): Viewable object.
        filename (str | list | tuple): Input filename(s).
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        NGLWidget: Viewable object.
    """
    from nglview import TextStructure
    # Atoms
    atom, pos = read_xyz(filename)
    atom, pos, pos_fod = split_fods(atom, pos, elec_symbols)
    view.add_component(TextStructure(create_pdb_str(atom, pos)))
    view[0].clear()
    view[0].add_ball_and_stick()
    # Spin up FODs
    if len(pos_fod[0]) > 0:
        view.add_component(TextStructure(create_pdb_str([elec_symbols[0]] * len(pos_fod[0]),
                           pos_fod[0])))
        view[1].clear()
        view[1].add_ball_and_stick(f'_{elec_symbols[0]}', color='red', radius=0.1)
    # Spin down FODs
    if len(pos_fod[1]) > 0:
        view.add_component(TextStructure(create_pdb_str([elec_symbols[1]] * len(pos_fod[1]),
                           pos_fod[1])))
        view[2].clear()
        view[2].add_ball_and_stick(f'_{elec_symbols[1]}', color='green', radius=0.1)
    view.center()
    return view


def _traj_view(view, filename):
    """Modify the view for a given TRAJ file.

    Args:
        view (NGLWidget): Viewable object.
        filename (str | list | tuple): Input filename(s).

    Returns:
        NGLWidget: Viewable object.
    """
    try:
        from nglview.base_adaptor import Structure, Trajectory
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise

    class eminusTrajectory(Trajectory, Structure):
        """eminusTrajectory object to handle trajectory files.

        The interface replicates the Trajectory classes in
        https://nglviewer.org/nglview/latest/_modules/nglview/adaptor.html

        Args:
            filenames (str | list | tuple): XYZ input file paths/names.
        """
        def __init__(self, filenames):
            """Initialize the eminusTrajectory object."""
            self.atoms = []
            if isinstance(filenames, str) and filenames.endswith(('.trj', '.traj')):
                trajectory = read_traj(filenames)
                for frame in trajectory:
                    self.atoms.append(Atoms(*frame))
            else:
                if isinstance(filenames, str) and filenames.endswith('.xyz'):
                    filenames = [filenames]
                for f in filenames:
                    self.atoms.append(Atoms(*read_xyz(f)))
            self.ext = 'pdb'
            self.params = {}
            self.id = str(uuid.uuid4())

        def get_coordinates(self, index):
            """Get the atom coordinates for a given frame.

            Args:get_Efermi
                index (int): Frame number.

            Returns:
                ndarray: Atom positions in Bohr.
            """
            return self.atoms[index].pos

        @property
        def n_frames(self):
            """Number of frames."""
            return len(self.atoms)

        def get_structure_string(self, index=0):
            """Get the structure string per frame in the PDB format.

            Keyword Args:
                index (int): Frame number.

            Returns:
                str: Structure string.
            """
            return create_pdb_str(self.atoms[index].atom, self.atoms[index].pos)

    trajectory = eminusTrajectory(filename)
    view.add_trajectory(trajectory)
    view.center()
    return view


def plot_bandstructure(scf, spin=0, size=(800, 600)):
    """Plot band structures.

    Reference: https://plotly.com/python

    Args:
        scf: SCF object.

    Keyword Args:
        spin (int): Spin index.
        size (tuple): Widget size.
    """
    try:
        import plotly.graph_objects as go
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise
    k_axis, special, label = kpoints2axis(scf.kpts)
    # Replace 'G' with the Greek 'Gamma' encoded in unicode
    label = ['\u0393' if l == 'G' else l for l in label]
    e_occ = ha2ev(get_epsilon(scf, scf.W, **scf._precomputed))

    if hasattr(scf, 'Z') or scf.atoms.occ.bands <= scf.atoms.occ.Nstate:
        Efermi = ha2ev(get_Efermi(scf))
    else:
        Efermi = find_Efermi(scf.atoms.occ, e_occ)

    fig = go.Figure()
    # Plot occupied bands
    for i in range(scf.atoms.occ.Nstate):
        fig.add_trace(go.Scatter(x=k_axis, y=e_occ[:, spin, i] - Efermi,
                      mode='lines+markers',
                      name=f'Band {i + 1}'))

    # Calculate and plot unoccupied bands if available
    if hasattr(scf, 'Z'):
        e_unocc = ha2ev(get_epsilon_unocc(scf, scf.W, scf.Z, **scf._precomputed))
        for i in range(scf.atoms.occ.Nempty):
            fig.add_trace(go.Scatter(x=k_axis, y=e_unocc[:, spin, i] - Efermi,
                          mode='lines+markers',
                          line={'dash': 'dash'},
                          name=f'Unocc. band {i + 1}'))

    fig.update_layout(
        width=size[0],
        height=size[1],
        showlegend=False,
        font={'size': 20},
        xaxis={'zeroline': False, 'showline': True, 'mirror': True, 'ticks': 'outside',
               'tickmode': 'array', 'tickvals': special, 'ticktext': label,
               'gridcolor': 'grey', 'gridwidth': 2},
        yaxis={'zeroline': False, 'showline': True, 'mirror': True, 'ticks': 'outside',
               'showgrid': False},
        xaxis_range=(0, k_axis[-1]),
        xaxis_title='k-path',
        yaxis_title='E - E<sub>F</sub> [eV]',
        hoverlabel_bgcolor='black',
        template='none')
    if executed_in_notebook():
        return fig
    return fig.show()


def view_kpts(kpts, path=True, special=True, connect=False, size=(600, 600)):
    """Display KPoints objects including Brillouin zones, special- and k-points.

    Reference: https://plotly.com/python

    Args:
        kpts: KPoints object.

    Keyword Args:
        path (bool): Determines if k-points should be plotted.
        special (bool): Determines if special points should be plotted.
        connect (bool): Determines if k-points should connected.
        size (tuple): Widget size.

    Returns:
        Figure: Viewable object.
    """
    try:
        import plotly.graph_objects as go
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise

    fig = go.Figure()
    # Calculate vertices of the Brillouin zone and plot them
    bz = get_brillouin_zone(kpts.a)
    for xx in bz:
        bz_data = go.Scatter3d(x=xx[:, 0], y=xx[:, 1], z=xx[:, 2],
                               name='Brillouin zone',
                               showlegend=False,
                               marker={'size': 0.1, 'color': 'black'})
        fig.add_trace(bz_data)

    # Plot special points if desired
    if special:
        for label, k_scaled in SPECIAL_POINTS[kpts.lattice].items():
            if label == 'G':
                label = '\u0393'  # noqa: PLW2901
            v = kpoint_convert(k_scaled, kpts.a)
            extra_data = go.Scatter3d(x=[v[0]], y=[v[1]], z=[v[2]],
                                      name=label,
                                      mode='markers',
                                      marker={'size': 4, 'opacity': 0.9})
            fig.add_trace(extra_data)

    # Plot optional k-points
    if connect:
        mode = 'lines+markers'
    else:
        mode = 'markers'
    if path:
        extra_data = go.Scatter3d(x=kpts.k[:, 0], y=kpts.k[:, 1], z=kpts.k[:, 2],
                                  name='k-points',
                                  mode=mode,
                                  marker={'size': 2, 'color': '#1a962b', 'opacity': 0.75})
        fig.add_trace(extra_data)

    # Theming
    scene = {'xaxis': {'title': 'b<sub>1</sub>'},
             'yaxis': {'title': 'b<sub>2</sub>'},
             'zaxis': {'title': 'b<sub>3</sub>'},
             'aspectmode': 'cube'}
    fig.update_layout(
        scene=scene,
        width=size[0],
        height=size[1],
        legend={'itemsizing': 'constant', 'title': 'Selection'},
        hoverlabel_bgcolor='black',
        template='none')
    if executed_in_notebook():
        return fig
    return fig.show()



#!/usr/bin/env python3
"""Symmetrize k-points."""

from ..logger import log


def symmetrize(atoms, space_group=False, time_reversal=True):
    """Symmetrize k-points of an Atoms object.

    Reference: WIREs Comput. Mol. Sci. 8, e1340.

    Args:
        atoms: Atoms object.

    Keyword Args:
        space_group (bool): Whether to consider space group symmetry.
        time_reversal (bool): Whether to consider time reversal symmetry.
    """
    try:
        from pyscf.pbc.gto import Cell
        from pyscf.pbc.lib.kpts import KPoints
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[symmetry]".\n\n')
        raise
    if not atoms.kpts.is_built:
        atoms.kpts.build()

    # Build Cell and KPoints objects
    cell = Cell()
    cell.unit = 'bohr'
    cell.atom = [[atoms.atom[ia], atoms.pos[ia]] for ia in range(atoms.Natoms)]
    cell.a = atoms.a
    cell.build()
    kpts = KPoints(cell, atoms.kpts.k)
    kpts.build(space_group_symmetry=space_group, time_reversal_symmetry=time_reversal)

    # Set the k-points
    atoms.kpts.k = kpts.kpts_ibz
    atoms.kpts.wk = kpts.weights_ibz
    atoms.kpts._k_scaled = None  # Remove the scaled k-points
    atoms.kpts.is_built = True  # Indicate the object as built
    atoms.is_built = False  # wk has to be passed to the occ object and G-vectors need to be updated



#!/usr/bin/env python3
"""Collection of constants and unit conversion functions.

For more about atomic units, see: https://en.wikipedia.org/wiki/Hartree_atomic_units
All values are directly calculated from the CODATA 2018 constants as found in SciPy.
"""
import numpy as np

#: Hartree in electronvolt.
# scipy.constants.value('Hartree energy in eV')
electronvolt = eV = 27.211386245988
#: Hartree in kcal per mol.
# scipy.constants.value('Hartree energy') * scipy.constants.value('Avogadro constant') / 4.184 / 1e3
kcalmol = 627.5094740630557
#: Bohr radius in Angstrom.
# scipy.constants.value('Bohr radius') * 1e10
Angstrom = A = 0.529177210903
#: Elementary charge times bohr radius in Debye.
# scipy.constants.value('atomic unit of electric dipole mom.') / \
# scipy.constants.value('hertz-inverse meter relationship') * 1e21
Debye = D = 2.5417464731818566


def ha2ev(E):
    """Convert Hartree to electronvolt.

    Args:
        E (float | ndarray): Energy in Hartree.

    Returns:
        float | ndarray: Energy in electronvolt.
    """
    return E * electronvolt


def ev2ha(E):
    """Convert electronvolt to Hartree.

    Args:
        E (float | ndarray): Energy in electronvolt.

    Returns:
        float | ndarray: Energy in Hartree.
    """
    return E / electronvolt


def ha2kcalmol(E):
    """Convert Hartree to kcal/mol.

    Args:
        E (float | ndarray): Energy in Hartree.

    Returns:
        float | ndarray: Energy in kcal/mol.
    """
    return E * kcalmol


def kcalmol2ha(E):
    """Convert kcal/mol to Hartree.

    Args:
        E (float | ndarray): Energy in kcal/mol.

    Returns:
        float | ndarray: Energy in Hartree.
    """
    return E / kcalmol


def ha2ry(E):
    """Convert Hartree to Rydberg.

    Args:
        E (float | ndarray): Energy in Hartree.

    Returns:
        float | ndarray: Energy in Rydberg.
    """
    return 2 * E


def ry2ha(E):
    """Convert Rydberg to Hartree.

    Args:
        E (float | ndarray): Energy in Rydberg.

    Returns:
        float | ndarray: Energy in Hartree.
    """
    return E / 2


def ang2bohr(r):
    """Convert Angstrom to Bohr.

    Args:
        r (float | ndarray): Length in Angstrom.

    Returns:
        float | ndarray: Length in Bohr.
    """
    return r / Angstrom


def bohr2ang(r):
    """Convert Bohr to Angstrom.

    Args:
        r (float | ndarray): Length in Bohr.

    Returns:
        float | ndarray: Length in Angstrom.
    """
    return r * Angstrom


def ebohr2d(p):
    """Convert e * Bohr to Debye.

    Args:
        p (float | ndarray): Electric dipole moment in e * Bohr.

    Returns:
        float | ndarray: Electric dipole moment in Debye.
    """
    return p * Debye


def d2ebohr(p):
    """Convert Debye to e * Bohr.

    Args:
        p (float | ndarray): Electric dipole moment in Debye.

    Returns:
        float | ndarray: Electric dipole moment in e * Bohr.
    """
    return p / Debye


def rad2deg(a):
    """Convert Radians to Degree.

    Args:
        a (float | ndarray): Angle in Radians.

    Returns:
        float | ndarray: Angle in Degree.
    """
    return a * 180 / np.pi


def deg2rad(a):
    """Convert Degree to Radians.

    Args:
        a (float | ndarray): Angle in Degree.

    Returns:
        float | ndarray: Angle in Radians.
    """
    return a * np.pi / 180



#!/usr/bin/env python3
"""TRAJ file handling."""
import numpy as np

from ..logger import log
from ..units import ang2bohr
from .xyz import write_xyz


def read_traj(filename):
    """Load atom species and positions from TRAJ files.

    TRAJ files are just multiple XYZ files appended to one file.
    See :func:`~eminus.io.xyz.read_xyz` for more information about the XYZ file format.

    Args:
        filename (str): TRAJ input file path/name.

    Returns:
        list[tuple[list, ndarray]]: Atom species and positions.
    """
    if not filename.endswith(('.trj', '.traj')):
        filename += '.traj'

    with open(filename, encoding='utf-8') as fh:
        lines = fh.readlines()
        Nlines = len(lines)

        # The first line contains the number of atoms
        Natoms = int(lines[0].strip())

        # The second line can contain a comment, print it if available
        comment = lines[1].strip()
        if comment:
            log.info(f'TRAJ file comment: "{comment}"')

        traj = []
        for frame in range(Nlines // (2 + Natoms)):
            atom = []
            pos = []
            # Following lines contain atom positions with the format: Atom x-pos y-pos z-pos
            for line in lines[(2 + Natoms) * frame + 2:(2 + Natoms) * (frame + 1)]:
                line_split = line.strip().split()
                atom.append(line_split[0])
                pos.append(np.float64(line_split[1:4]))
            # XYZ files are in Angstrom, so convert to Bohr
            pos = ang2bohr(np.asarray(pos))
            traj.append((atom, pos))
    return traj


def write_traj(obj, filename, fods=None, elec_symbols=('X', 'He')):
    """Generate TRAJ files from atoms objects.

    TRAJ files are just multiple XYZ files appended to one file.
    See :func:`~eminus.io.xyz.write_xyz` for more information about the XYZ file format.

    Args:
        obj: Atoms or SCF object or list/tuple of these objects.
        filename (str): TRAJ output file path/name.

    Keyword Args:
        fods (list | None): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        None.
    """
    if not filename.endswith(('.trj', '.traj')):
        filename += '.traj'

    if isinstance(obj, (list, tuple)):
        for iobj in obj:
            write_xyz(iobj, filename, fods=fods, elec_symbols=elec_symbols, trajectory=True)
    else:
        write_xyz(obj, filename, fods=fods, elec_symbols=elec_symbols, trajectory=True)



#!/usr/bin/env python3
"""Determine occupations for atomic systems from simple inputs."""
import dataclasses
import numbers

import numpy as np

from .logger import log
from .tools import fermi_distribution


@dataclasses.dataclass
class Occupations:
    """Occupations class to save electronic state information in one place.

    The attribute Nelec has to be given first after instantiation.
    """
    # Set the private variable for the attributes that are properties.
    _Nelec: int = 0          #: Number of electrons.
    _Nspin: int = 0          #: Number of spin states.
    _spin: int = 0           #: Number of unpaired electrons.
    _charge: int = 0         #: System charge.
    _Nstate: int = 0         #: Number of states.
    _Nempty: int = 0         #: Number of empty states.
    _Nk: int = 1             #: Number of k-points.
    _bands: int = 0          #: Number of bands.
    _smearing: float = 0     #: Smearing width.
    is_filled: bool = False  #: Determines the Occupations object fill status.

    # ### Class properties ###

    @property
    def Nelec(self):
        """Number of electrons."""
        return self._Nelec

    @Nelec.setter
    def Nelec(self, value):
        # Only update Nelec if it actually gets updated
        if self._Nelec != int(value):
            self._Nelec = int(value)
            self.is_filled = False

    @property
    def Nspin(self):
        """Number of spin states."""
        return self._Nspin

    @Nspin.setter
    def Nspin(self, value):
        # Use a spin-paired calculation for an even number of electrons
        if value is None:
            if self.Nelec % 2 == 0 and self.spin == 0:
                value = 1
            else:
                value = 2
        # Only update if needed
        if self._Nspin != int(value):
            self._Nspin = int(value)
            self.spin = self.spin
            self.is_filled = False
            if hasattr(self, 'f'):
                log.warning('Reset previously set fillings.')

    @property
    def spin(self):
        """Number of unpaired electrons."""
        return self._spin

    @spin.setter
    def spin(self, value):
        # If no spin is given try to determine it from the number of electrons
        if value is None:
            if self.Nelec % 2 == 0:
                value = 0
            else:
                value = 1
        if self._spin != int(value):
            self.is_filled = False
        # We have no spin in the spin-paired case
        if self.Nspin == 1:
            self._spin = 0
        else:
            self._spin = int(value)

    @property
    def charge(self):
        """System charge."""
        return self._charge

    @charge.setter
    def charge(self, value):
        # If we set a charge via this setter update the number of electrons
        if hasattr(self, 'Nelec'):
            self.Nelec += self._charge - int(value)
        if self._charge != int(value):
            self._charge = int(value)
            self.is_filled = False

    @property
    def f(self):
        """Occupation numbers per state."""
        return self._f

    @f.setter
    def f(self, value):
        # Make sure the occupations are in a two-dimensional array
        if isinstance(value, (list, tuple, np.ndarray)):
            value = np.atleast_2d(value)
        self.is_filled = False
        # This setter will only be called when explicitly setting f
        # Call the fill function in that case
        self.fill(value)

    @property
    def Nempty(self):
        """Number of empty states."""
        return self._Nempty

    @Nempty.setter
    def Nempty(self, value):
        # Do not overwrite the fillings when setting a negative amount of empty states, let the user
        # handle this
        if value < 0:
            value = 0
            log.warning('Negative empty states are not allowed. Try to set the fillings manually.')
        self._Nempty = int(value)

    @property
    def Nk(self):
        """Number of k-points."""
        return self._Nk

    @Nk.setter
    def Nk(self, value):
        self._Nk = int(value)
        self.is_filled = False

    @property
    def wk(self):
        """k-point weights."""
        return self._wk

    @wk.setter
    def wk(self, value):
        self._wk = np.asarray(value)
        self._Nk = len(self._wk)
        self.is_filled = False

    @property
    def bands(self):
        """Total number of bands."""
        return self._bands

    @bands.setter
    def bands(self, value):
        self._bands = int(value)
        self.is_filled = False

    @property
    def smearing(self):
        """Smearing width."""
        return self._smearing

    @smearing.setter
    def smearing(self, value):
        self._smearing = value
        self.is_filled = False
        if self.Nempty > 0:
            log.warning('Empty states with smearing enabled found.')

        if value > 0:
            log.warning('The current smearing implementation is untested and should not be used!')

    # ### Read-only properties ###

    @property
    def multiplicity(self):
        """Multiplicity, i.e., 2S+1."""
        return self.spin + 1

    @property
    def Nstate(self):
        """Number of states."""
        return self._Nstate

    @property
    def F(self):
        """Diagonal matrices of f per k-point and spin."""
        return [[np.diag(f) for f in f_spin] for f_spin in self.f]

    # ### Class methods ###

    def fill(self, f=None):
        """Fill the states of the object.

        Keyword Args:
            f: (float | ndarray | None): Fillings.
        """
        # Do nothing if the object is already filled
        if self.is_filled:
            return self
        # If no f is given just use the standard fillings: 2 for restricted and 1 for unrestricted
        if f is None:
            f = 2 / self.Nspin
        self._update_from_fillings(f)
        # Assure that no electrons have been lost
        if np.sum(self.f) != self.Nelec:
            ValueError(f'Sum of fillings ({np.sum(self.f[0])}) differs from Nelec ({self.Nelec}).')
        self.is_filled = True
        return self

    kernel = fill

    def _update_from_fillings(self, value):
        """Update fillings."""
        # Do not use the setter methods in this place to not trigger the setter effects
        # If f is a number use this occupation for all states
        if isinstance(value, numbers.Real):
            # Do not leave the states array empty when no electrons are present
            if self.Nelec <= 0:
                self._Nstate = 1
                self._f = np.zeros((self.Nk, self.Nspin, 1))
            elif self.Nspin == 1 or self.Nelec % 2 == self.spin % 2:
                self._integer_fillings(value)
            else:
                log.warning('Using fractional occupations.')
                self._fractional_fillings(value)
        # If f is an array redetermine all attributes from it
        else:
            self._f = value
            self._Nspin = len(value)
            self._Nstate = len(value[0])
            self._charge += self.Nelec - int(np.sum(value))
            self.Nelec = int(np.sum(value))
            if self.Nspin == 1:
                self.spin = 0
            else:
                self.spin = abs(np.sum(self.f[0] - self.f[1]))

    def _integer_fillings(self, f):
        """Update fillings while maintaining integer occupations numbers.

        Args:
            f: (int | ndarray): Fillings.
        """
        # Determine the electrons per spin channel
        Nup = self.Nelec // self.Nspin + self.spin // self.Nspin + self.Nelec % self.Nspin
        Ndw = self.Nelec // self.Nspin - self.spin // self.Nspin
        elecs = np.array([Nup, Ndw])

        # Get the number of states
        Nstate = int(np.ceil(max(elecs / f)))
        # If no bands are set, set them now
        if self.bands == 0:
            self.bands = Nstate
            self._Nstate = Nstate
        if self.bands < Nstate:
            log.error('Number of bands is smaller than the number of valence electrons.')
        # Disallow empty bands when using smearing
        if self.smearing > 0:
            self._Nstate = self.bands
        # Set the number of empty states if no smearing is used
        else:
            self._Nstate = Nstate
            self.Nempty = self.bands - Nstate

        # Simply build the occupations array
        self._f = f * np.ones((self.Nspin, self._Nstate), dtype=int)
        # If we have filled too much correct it in the second spin channel
        # The following procedure ensures that we have no negative fillings
        rest = np.sum(self._f) - self.Nelec
        i = 1
        while rest > 0:
            # If the occupation is greater than the rest we are done with removing electrons
            if self._f[-1, -i] >= rest:
                self._f[-1, -i] -= rest
                break
            # Otherwise zero out the state, update the rest, and move to the next state
            rest -= self._f[-1, -i]
            self._f[-1, -i] = 0
            i += 1
        self._f = np.vstack([[self._f]] * self.Nk)

    def _fractional_fillings(self, f):
        """Update fillings while allowing fractional occupation numbers.

        Args:
            f: (float | ndarray): Fillings.
        """
        # Determine the electrons per spin channel
        Nup = self.Nelec / self.Nspin + self.spin / self.Nspin
        Ndw = self.Nelec / self.Nspin - self.spin / self.Nspin
        elecs = np.array([Nup, Ndw])

        # Get the number of states
        Nstate = int(np.ceil(max(elecs / f)))
        # If no bands are set, set them now
        if self.bands == 0:
            self.bands = Nstate
            self._Nstate = Nstate
        if self.bands < Nstate:
            log.error('Number of bands is smaller than the number of valence electrons.')
        # Disallow empty bands when using smearing
        if self.smearing > 0:
            self._Nstate = self.bands
        # Set the number of empty states if no smearing is used
        else:
            self._Nstate = Nstate
            self.Nempty = self.bands - Nstate

        # Simply build the occupations array
        self._f = f * np.ones((self.Nspin, self._Nstate))
        # If we have filled too much correct it in both spin channels
        # The following procedure ensures that we have no negative fillings
        rest = np.sum(self._f, axis=1) - elecs
        for s in range(self.Nspin):
            i = 1
            while rest[s] > 0:
                # If the occupation is greater than the rest we are done with removing electrons
                if self._f[s, -i] >= rest[s]:
                    self._f[s, -i] -= rest[s]
                    break
                # Otherwise zero out the state, update the rest, and move to the next state
                rest[s] -= self._f[s, -i]
                self._f[s, -i] = 0
                i += 1
        self._f = np.vstack([[self._f]] * self.Nk)

    def __repr__(self):
        """Print the parameters stored in the Occupations object."""
        return f'Spin handling: {"un" if self.Nspin == 2 else ""}restricted\n' \
               f'Number of electrons: {self.Nelec}\n' \
               f'Spin: {self.spin}\n' \
               f'Charge: {self.charge}\n' \
               f'Number of bands: {self.bands}\n' \
               f'Number of states: {self.Nstate}\n' \
               f'Number of empty states: {self.Nempty}\n' \
               f'Number of k-points: {self.Nk}\n' \
               f'Smearing width: {self.smearing}\n' \
               f'Fillings: \n{self.f if self.is_filled else "Not filled"}'

    def smear(self, epsilon):
        """Update fillings according to a Fermi distribution.

        Args:
            epsilon (ndarray): Eigenenergies.
            Efermi (float): Fermi energy.
        """
        if self.smearing == 0:
            log.info('Smearing is set to zero, nothing to do.')
            return 0

        Efermi = find_Efermi(self, epsilon)

        f = np.empty_like(self.f)
        for ik in range(self.Nk):
            for spin in range(self.Nspin):
                for istate in range(self.Nstate):
                    f[ik, spin, istate] = fermi_distribution(epsilon[ik, spin, istate], Efermi,
                                                             self.smearing)
        self._f = f * 2 / self.Nspin
        return Efermi


def sumkg(occ, epsilon, Efermi):
    """Sum occupation numbers for given eigenenergies up to the Fermi energy.

    Reference: https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/occupations.jl

    Args:
        occ: Occupations object.
        epsilon (ndarray): Eigenenergies.
        Efermi (float): Fermi energy.

    Returns:
        float: Summed occupations.
    """
    wk = occ.wk
    if occ.Nspin == 1:
        wk = 2 * wk

    ss = 0
    for ik in range(len(wk)):
        for spin in range(occ.Nspin):
            ss1 = 0
            for istate in range(occ.Nstate):
                ss1 = ss1 + fermi_distribution(epsilon[ik, spin, istate], Efermi, occ.smearing)
            ss += wk[ik] * ss1
    return ss


def find_Efermi(occ, epsilon):
    """Find the Fermi energy for given eigenenergies.

    Reference: https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/occupations.jl

    Args:
        occ: Occupations object.
        epsilon (ndarray): Eigenenergies.

    Returns:
        float: Fermi energy.
    """
    eps = 1e-10

    Elw = np.min(epsilon[..., 0])
    Eup = np.max(epsilon[..., -1])
    Elw -= 2 * occ.smearing
    Eup += 2 * occ.smearing

    sumklw = sumkg(occ, epsilon, Elw)
    sumkup = sumkg(occ, epsilon, Eup)
    if (sumkup - occ.Nelec) < -1e-15 or (sumklw - occ.Nelec) > 1e-15:
        log.exception('Bounds for Efermi not found.')

    Ef = 0.5 * (Eup + Elw)
    Ef_old = Ef
    for _ in range(500):
        sumkmid = sumkg(occ, epsilon, Ef)
        if abs(sumkmid - occ.Nelec) < eps:
            return Ef
        if sumkmid - occ.Nelec < -eps:
            Elw = Ef
        else:
            Eup = Ef
        Ef = 0.5 * (Eup + Elw)
        diff_Ef = abs(Ef - Ef_old)
        if diff_Ef < eps:
            return Ef
        Ef_old = Ef
    return Ef



#!/usr/bin/env python3
"""Generate k-points and sample band paths."""
import numbers

import numpy as np
from scipy.linalg import inv, norm
from scipy.spatial import Voronoi

from .data import LATTICE_VECTORS, SPECIAL_POINTS
from .logger import log


class KPoints:
    """KPoints object that holds k-points properties and build methods.

    Args:
        lattice (str): Lattice system.
        a (float | list | tuple | ndarray | None): Cell size.
    """
    def __init__(self, lattice, a=None):
        """Initialize the KPoints object."""
        self.lattice = lattice        #: Lattice system.
        if a is None:
            a = LATTICE_VECTORS[self.lattice]
        if isinstance(a, numbers.Real):
            a = a * np.asarray(LATTICE_VECTORS[self.lattice])
        self.a = a                   #: Cell size.
        self.kmesh = [1, 1, 1]       #: k-point mesh.
        self.wk = [1]                #: k-point weights.
        self.k = [[0, 0, 0]]         #: k-point coordinates.
        self.kshift = [0, 0, 0]      #: k-point shift-vector.
        self._gamma_centered = True  #: Generate a Gamma-point centered grid.
        self.is_built = True         #: Determines the KPoints object build status.

    # ### Class properties ###

    @property
    def kmesh(self):
        """Monkhorst-Pack k-point mesh."""
        return self._kmesh

    @kmesh.setter
    def kmesh(self, value):
        if value is not None:
            if isinstance(value, numbers.Integral):
                value = value * np.ones(3, dtype=int)
            self._kmesh = np.asarray(value)
            self.path = None
            self.is_built = False
        # If we set a band path to the object the k-mesh gets set to None
        else:
            self._kmesh = None

    @property
    def wk(self):
        """k-point weights."""
        return self._wk

    @wk.setter
    def wk(self, value):
        self._wk = np.asarray(value)
        self._Nk = len(self._wk)
        self.is_built = False

    @property
    def k(self):
        """k-point coordinates."""
        return self._k

    @k.setter
    def k(self, value):
        self._k = np.asarray(value)
        self.is_built = False

    @property
    def Nk(self):
        """Number of k-points."""
        return self._Nk

    @Nk.setter
    def Nk(self, value):
        self._Nk = int(value)
        self.is_built = False

    @property
    def kshift(self):
        """k-point shift-vector."""
        return self._kshift

    @kshift.setter
    def kshift(self, value):
        self._kshift = np.asarray(value)
        self.is_built = False

    @property
    def gamma_centered(self):
        """Generate a Gamma-point centered grid."""
        return self._gamma_centered

    @gamma_centered.setter
    def gamma_centered(self, value):
        if value != self._gamma_centered:
            self._gamma_centered = value
            self.is_built = False

    @property
    def path(self):
        """k-point band path."""
        return self._path

    @path.setter
    def path(self, value):
        if value is not None:
            self._path = value.upper()
            self.kmesh = None
            self.is_built = False
        # If we set a k-mesh to the object the band path gets set to None
        else:
            self._path = None

    # ### Read-only properties ###

    @property
    def k_scaled(self):
        """Scaled k-point coordinates."""
        # This will not be set when setting the k-point coordinates manually
        return self._k_scaled

    # ### Class methods ###

    def build(self):
        """Build all parameters of the KPoints object."""
        if self.lattice == 'sc' and not (self.a == np.diag(np.diag(self.a))).all():
            log.warning('Lattice system and lattice vectors do not match.')
        if self.is_built:
            return self
        if self.kmesh is not None:
            if self.gamma_centered:
                self._k_scaled = gamma_centered(self.kmesh)
            else:
                self._k_scaled = monkhorst_pack(self.kmesh)
        else:
            self._k_scaled = bandpath(self)
        # Without removing redundancies the weight is the same for all k-points
        self.wk = np.ones(len(self._k_scaled)) / len(self._k_scaled)
        self.k = kpoint_convert(self._k_scaled, self.a) + self.kshift
        self.is_built = True
        return self

    kernel = build

    def _assert_gamma_only(self):
        """Make sure that the object only contains the Gamma point."""
        if not np.all(self.k == 0):
            log.exception('The k-points object does not contain only the Gamma point.')
            raise NotImplementedError

    def __repr__(self):
        """Print the parameters stored in the KPoints object."""
        return f'Number of k-points: {self.Nk}\n' \
               f'k-mesh: {self.kmesh}\n' \
               f'Band path: {self.path}\n' \
               f'Shift: {self.kshift}\n' \
               f'Weights: {self.wk}'


def kpoint_convert(k_points, lattice_vectors):
    """Convert scaled k-points to cartesian coordinates.

    Reference: https://gitlab.com/ase/ase/-/blob/master/ase/dft/kpoints.py

    Args:
        k_points (ndarray): k-points.
        lattice_vectors (ndarray): Lattice vectors.

    Returns:
        ndarray: k-points in cartesian coordinates.
    """
    inv_cell = 2 * np.pi * inv(lattice_vectors).T
    return k_points @ inv_cell


def monkhorst_pack(nk):
    """Generate a Monkhorst-Pack mesh of k-points, i.e., equally spaced k-points.

    Reference: https://gitlab.com/ase/ase/-/blob/master/ase/dft/kpoints.py

    Args:
        nk (list | tuple | ndarray): Number of k-points per axis.

    Returns:
        ndarray: k-points.
    """
    # Same index matrix as in Atoms._get_index_matrices()
    M = np.indices(nk).transpose((1, 2, 3, 0)).reshape((-1, 3))
    return (M + 0.5) / nk - 0.5  # Normal Monkhorst-Pack grid


def gamma_centered(nk):
    """Generate a Gamma-point centered mesh of k-points.

    Reference: https://github.com/pyscf/pyscf/blob/master/pyscf/pbc/gto/cell.py

    Args:
        nk (list | tuple | ndarray): Number of k-points per axis.

    Returns:
        ndarray: k-points.
    """
    # Same index matrix as in Atoms._get_index_matrices()
    M = np.indices(nk).transpose((1, 2, 3, 0)).reshape((-1, 3))
    return M / nk  # Gamma-centered grid


def bandpath(kpts):
    """Generate sampled band paths.

    Args:
        kpts: KPoints object.

    Returns:
        ndarray: Sampled k-points.
    """
    # Convert path to a list and get special points
    path_list = list(kpts.path)
    s_points = SPECIAL_POINTS[kpts.lattice]
    # Commas indicate jumps and are no special points
    N_special = len([p for p in path_list if p != ','])

    # Input handling
    N = kpts.Nk
    if N_special > N:
        log.warning('Sampling is smaller than the number of special points.')
        N = N_special
    for p in path_list:
        if p not in {*s_points, ','}:
            raise KeyError(f'{p} is not a special point for the {kpts.lattice} lattice.')

    # Calculate distances between special points
    dists = []
    for i in range(len(path_list) - 1):
        if ',' not in path_list[i:i + 2]:
            # Use subtract since s_points are lists
            dist = np.subtract(s_points[path_list[i + 1]], s_points[path_list[i]])
            dists.append(norm(kpoint_convert(dist, kpts.a)))
        else:
            # Set distance to zero when jumping between special points
            dists.append(0)

    # Calculate sample points between the special points
    scaled_dists = (N - N_special) * np.array(dists) / sum(dists)
    samplings = np.int64(np.round(scaled_dists))

    # If our sampling does not match the given N add the difference to the longest distance
    if N - N_special - np.sum(samplings) != 0:
        samplings[np.argmax(samplings)] += N - N_special - np.sum(samplings)

    # Generate k-point coordinates
    k_points = [s_points[path_list[0]]]  # Insert the first special point
    for i in range(len(path_list) - 1):
        # Only do something when not jumping between special points
        if ',' not in path_list[i:i + 2]:
            s_start = s_points[path_list[i]]
            s_end = s_points[path_list[i + 1]]
            # Get the vector between special points
            k_dist = np.subtract(s_end, s_start)
            # Add scaled vectors to the special point to get the new k-points
            k_points += [s_start + k_dist * (n + 1) / (samplings[i] + 1)
                         for n in range(samplings[i])]
            # Append the special point we are ending at
            k_points.append(s_end)
        # If we jump, add the new special point to start from
        elif path_list[i] == ',':
            k_points.append(s_points[path_list[i + 1]])
    return np.asarray(k_points)


def kpoints2axis(kpts):
    """Generate the x-axis for band structures and the respective band path.

    Args:
        kpts: KPoints object.

    Returns:
        tuple[ndarray, ndarray, list]: k-point axis, special point coordinates, and labels.
    """
    # Convert path to a list and get the special points
    path_list = list(kpts.path)
    s_points = SPECIAL_POINTS[kpts.lattice]

    # Calculate the distances between k-points
    k_dist = kpts.k_scaled[1:] - kpts.k_scaled[:-1]
    dists = norm(kpoint_convert(k_dist, kpts.a), axis=1)

    # Create the labels
    labels = []
    for i in range(len(path_list)):
        # If a jump happened before the current step the special point is already included
        if i > 1 and path_list[i - 1] == ',':
            continue
        # Append the special point if no jump happens
        if ',' not in path_list[i:i + 2]:
            labels.append(path_list[i])
        # When jumping join the special points to one label
        elif path_list[i] == ',':
            labels.append(''.join(path_list[i - 1:i + 2]))

    # Get the indices of the special points
    special_indices = [0]  # The first special point is trivial
    for p in labels[1:]:
        # Only search the k-points starting from the previous special point
        shift = special_indices[-1]
        k = kpts.k_scaled[shift:]
        # We index p[0] since p could be a joined label of a jump
        # This expression simply finds the special point in the k_points matrix
        index = np.flatnonzero((k == s_points[p[0]]).all(axis=1))[0] + shift
        special_indices.append(index)
        # Set the distance between special points to zero if we have a jump
        if ',' in p:
            dists[index] = 0

    # Insert a zero at the beginning and add up the lengths to create the k-axis
    k_axis = np.append([0], np.cumsum(dists))
    return k_axis, k_axis[special_indices], labels


def get_brillouin_zone(lattice_vectors):
    """Generate the Brillouin zone for given lattice vectors.

    The Brillouin zone can be constructed with a Voronoi decomposition of the reciprocal lattice.

    Reference: http://staff.ustc.edu.cn/~zqj/posts/howto-plot-brillouin-zone

    Args:
        lattice_vectors (ndarray): Lattice vectors.

    Returns:
        ndarray: Brillouin zone vertices.
    """
    inv_cell = kpoint_convert(np.eye(3), lattice_vectors)

    px, py, pz = np.tensordot(inv_cell, np.mgrid[-1:2, -1:2, -1:2], axes=(0, 0))
    points = np.c_[px.ravel(), py.ravel(), pz.ravel()]
    vor = Voronoi(points)

    bz_ridges = []
    for pid, rid in zip(vor.ridge_points, vor.ridge_vertices):
        if pid[0] == 13 or pid[1] == 13:
            bz_ridges.append(vor.vertices[np.r_[rid, [rid[0]]]])
    return bz_ridges



#!/usr/bin/env python3
"""Basis set dependent operators for a plane wave basis.

These operators act on discretized wave functions, i.e., the arrays W.

These W are column vectors. This has been chosen to let theory and code coincide, e.g.,
W^dagger W becomes :code:`W.conj().T @ W`.

The downside is that the i-th state will be accessed with W[:, i] instead of W[i].
Choosing the i-th state makes the array 1d.

These operators can act on six different options, namely

1. the real-space
2. the real-space (1d)
3. the full reciprocal space
4. the full reciprocal space (1d)
5. the active reciprocal space
6. the active reciprocal space (1d)

The active space is the truncated reciprocal space by restricting it with a sphere given by ecut.

Every spin dependence will be handled with handle_spin_gracefully by calling the operators for each
spin individually. The same goes for the handling of k-points, while for k-points W is represented
as a list of arrays. This gives the final indexing for the k-point k, spin s, and state n of
W[ik][s, :, n].
"""
import copy

import numpy as np
from scipy.fft import fftn, ifftn

from . import config
from .utils import handle_k_gracefully, handle_k_indexable, handle_spin_gracefully, handle_torch


# Spin handling is trivial for this operator
@handle_k_gracefully
def O(atoms, W):
    """Overlap operator.

    This operator acts on the options 3, 4, 5, and 6.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    """
    return atoms.Omega * W


@handle_spin_gracefully
def L(atoms, W, ik=-1):
    """Laplacian operator with k-point dependency.

    This operator acts on options 3 and 5.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        ik (int): k-point index.

    Returns:
        ndarray: The operator applied on W.
    """
    # Gk2 is a normal 1d row vector, reshape it so it can be applied to the column vector W
    if len(W) == len(atoms.Gk2c[ik]):
        Gk2 = atoms.Gk2c[ik][:, None]
    else:
        Gk2 = atoms.Gk2[ik][:, None]
    return -atoms.Omega * Gk2 * W


@handle_spin_gracefully
def Linv(atoms, W):
    """Inverse Laplacian operator.

    This operator acts on options 3 and 4.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    """
    # Ignore the division by zero for the first elements
    with np.errstate(divide='ignore', invalid='ignore'):
        if W.ndim == 1:
            # One could do some proper indexing with [1:] but indexing is slow
            out = W / (atoms.G2 * -atoms.Omega)
            out[0] = 0
        else:
            # G2 is a normal 1d row vector, reshape it so it can be applied to the column vector W
            G2 = atoms.G2[:, None]
            out = W / (G2 * -atoms.Omega)
            out[0, :] = 0
    return out


@handle_torch
@handle_k_indexable
@handle_spin_gracefully
def I(atoms, W, ik=-1):
    """Backward transformation from reciprocal space to real-space.

    This operator acts on the options 3, 4, 5, and 6.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        ik (int): k-point index.

    Returns:
        ndarray: The operator applied on W.
    """
    n = atoms.Ns

    # If W is in the full space do nothing with W
    if len(W) == len(atoms.Gk2[ik]):
        Wfft = np.copy(W)
    else:
        # Fill with zeros if W is in the active space
        if W.ndim == 1:
            Wfft = np.zeros(n, dtype=W.dtype)
        else:
            Wfft = np.zeros((n, W.shape[-1]), dtype=W.dtype)
        Wfft[atoms.active[ik]] = W

    # `workers` sets the number of threads the FFT operates on
    # `overwrite_x` allows writing in Wfft, but since we do not need Wfft later on, we can set this
    # for a little bit of extra performance
    # Normally, we would have to multiply by n in the end for the correct normalization, but we can
    # ignore this step when properly setting the `norm` option for a faster operation
    if W.ndim == 1:
        Wfft = Wfft.reshape(atoms.s)
        Finv = ifftn(Wfft, workers=config.threads, overwrite_x=True, norm='forward').ravel()
    else:
        # Here we reshape the input like in the 1d case but add an extra dimension in the end,
        # holding the number of states
        Wfft = Wfft.reshape(np.append(atoms.s, W.shape[-1]))
        # Tell the function that the FFT only has to act on the first 3 axes
        Finv = ifftn(Wfft, workers=config.threads, overwrite_x=True, norm='forward',
                     axes=(0, 1, 2)).reshape((n, W.shape[-1]))
    return Finv


@handle_torch
@handle_k_indexable
@handle_spin_gracefully
def J(atoms, W, ik=-1, full=True):
    """Forward transformation from real-space to reciprocal space.

    This operator acts on options 1 and 2.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        ik (int): k-point index.
        full (bool): Whether to transform in the full or in the active space.

    Returns:
        ndarray: The operator applied on W.
    """
    n = atoms.Ns
    Wfft = np.copy(W)

    # `workers` sets the number of threads the FFT operates on
    # `overwrite_x` allows writing in Wfft, but since we do not need Wfft later on, we can set this
    # for a little bit of extra performance
    # Normally, we would have to divide by n in the end for the correct normalization, but we can
    # ignore this step when properly setting the `norm` option for a faster operation
    if W.ndim == 1:
        Wfft = Wfft.reshape(atoms.s)
        F = fftn(Wfft, workers=config.threads, overwrite_x=True, norm='forward').ravel()
    else:
        Wfft = Wfft.reshape(np.append(atoms.s, W.shape[-1]))
        F = fftn(Wfft, workers=config.threads, overwrite_x=True, norm='forward',
                 axes=(0, 1, 2)).reshape((n, W.shape[-1]))

    # There is no way to know if J has to transform to the full or the active space
    # but normally it transforms to the full space
    if not full:
        return F[atoms.active[ik]]
    return F


@handle_torch
@handle_k_indexable
@handle_spin_gracefully
def Idag(atoms, W, ik=-1, full=False):
    """Conjugated backward transformation from real-space to reciprocal space.

    This operator acts on options 1 and 2.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        ik (int): k-point index.
        full (bool): Whether to transform in the full or in the active space.

    Returns:
        ndarray: The operator applied on W.
    """
    n = atoms.Ns
    F = J(atoms, W, ik, full)
    return F * n


@handle_torch
@handle_k_indexable
@handle_spin_gracefully
def Jdag(atoms, W, ik=-1):
    """Conjugated forward transformation from reciprocal space to real-space.

    This operator acts on the options 3, 4, 5, and 6.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        ik (int): k-point index.

    Returns:
        ndarray: The operator applied on W.
    """
    n = atoms.Ns
    Finv = I(atoms, W, ik)
    return Finv / n


@handle_spin_gracefully
def K(atoms, W, ik):
    """Preconditioning operator with k-point dependency.

    This operator acts on options 3 and 5.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        ik (int): k-point index.

    Returns:
        ndarray: The operator applied on W.
    """
    # Gk2c is a normal 1d row vector, reshape it so it can be applied to the column vector W
    return W / (1 + atoms.Gk2c[ik][:, None])


def T(atoms, W, dr):
    """Translation operator.

    This operator acts on options 5 and 6.

    Reference: https://ccrma.stanford.edu/~jos/st/Shift_Theorem.html

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        ik (int): k-point index.
        dr (ndarray): Real-space shifting vector.

    Returns:
        ndarray: The operator applied on W.
    """
    if isinstance(W, np.ndarray) and W.ndim == 3:
        return np.asarray([T(atoms, Wspin, dr) for Wspin in W])

    if isinstance(W, np.ndarray):
        atoms.kpts._assert_gamma_only()
        if len(W) == len(atoms.Gk2c[0]):
            G = atoms.G[atoms.active[0]]
        elif len(W) == len(atoms.Gk2c[-1]):
            G = atoms.G[atoms.active[-1]]
        else:
            G = atoms.G
        factor = np.exp(-1j * G @ dr)
        if W.ndim == 2:
            factor = factor[:, None]
        return factor * W

    Wshift = copy.deepcopy(W)
    for ik in range(atoms.kpts.Nk):
        # Do the shift by multiplying a phase factor, given by the shift theorem
        if W[ik].shape[1] == len(atoms.Gk2c[ik]):
            Gk = atoms.G[atoms.active[ik]] + atoms.kpts.k[ik]
        else:
            Gk = atoms.G + atoms.kpts.k[ik]
        Wshift[ik] = np.exp(-1j * Gk @ dr)[:, None] * W[ik]
    return Wshift



#!/usr/bin/env python3
"""Minimization algorithms for fixed Hamiltonians.

Similar to :mod:`eminus.minimizer` but for a fixed Hamiltonian the implementation can be simplified
and made more performant.
"""
import copy
import logging

import numpy as np
from scipy.linalg import inv, sqrtm

from .dft import H, orth, orth_unocc
from .energies import get_Eband
from .logger import name
from .minimizer import cg_method, cg_test, check_convergence, linmin_test
from .utils import dotprod


def scf_step_occ(scf, W):
    """Perform one SCF step for an occupied band minimization calculation.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        float: Band energy.
    """
    atoms = scf.atoms
    scf.Y = orth(atoms, W)
    return get_Eband(scf, scf.Y, **scf._precomputed)


def scf_step_unocc(scf, Z):
    """Perform one SCF step for an unoccupied band minimization calculation.

    Args:
        scf: SCF object.
        Z (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        float: Band energy.
    """
    atoms = scf.atoms
    scf.D = orth_unocc(atoms, scf.Y, Z)
    return get_Eband(scf, scf.D, **scf._precomputed)


def get_grad_occ(scf, ik, spin, W, **kwargs):
    """Calculate the occupied band energy gradient with respect to W.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        ik (int): k-point index.
        spin (int): Spin variable to track whether to do the calculation for spin up or down.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        **kwargs: See :func:`H`.

    Returns:
        ndarray: Gradient.
    """
    atoms = scf.atoms
    W = orth(atoms, W)
    HW = H(scf, ik, spin, W, **kwargs)
    WHW = W[ik][spin].conj().T @ HW
    OW = atoms.O(W[ik][spin])
    U = W[ik][spin].conj().T @ OW
    invU = inv(U)
    U12 = sqrtm(invU)
    # grad E = (I - O(Y) Ydag) H(Y) U^-0.5
    return atoms.kpts.wk[ik] * ((HW - OW @ WHW) @ U12)


def get_grad_unocc(scf, ik, spin, Z, **kwargs):
    """Calculate the unoccupied band energy gradient with respect to Z.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        ik (int): k-point index.
        spin (int): Spin variable to track whether to do the calculation for spin up or down.
        Z (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        **kwargs: See :func:`H`.

    Returns:
        ndarray: Gradient.
    """
    atoms = scf.atoms
    Y = scf.Y[ik][spin][:, scf.atoms.occ.f[ik][spin] > 0]
    Ydag = Y.conj().T
    # We need X12 later, so orthogonalize in-place and only the current state
    rhoZ = Z[ik][spin] - Y @ Ydag @ atoms.O(Z[ik][spin])
    X12 = inv(sqrtm(rhoZ.conj().T @ atoms.O(rhoZ)))
    D = rhoZ @ X12
    # Create the correct input shape for the Hamiltonian
    D_tmp = [np.empty_like(Zk) for Zk in Z]
    D_tmp[ik][spin] = D
    HD = H(scf, ik, spin, D_tmp, **kwargs)
    DHD = D.conj().T @ HD
    I = np.eye(Z[ik].shape[1])
    # grad E = (I - O(Y) Ydag) (I - O(D) Ddag) H(D) X^-0.5
    return atoms.kpts.wk[ik] * ((I - atoms.O(Y) @ Ydag) @ (HD - atoms.O(D) @ DHD) @ X12)


@name('steepest descent minimization')
def sd(scf, W, Nit, cost=scf_step_occ, grad=get_grad_occ, condition=check_convergence, betat=1,
       **kwargs):
    """Steepest descent minimization algorithm for a fixed Hamiltonian.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[list, ndarray]: Band energies per SCF cycle and optimized expansion coefficients.
    """
    atoms = scf.atoms
    costs = []

    for _ in range(Nit):
        c = cost(scf, W)
        costs.append(c)
        if condition(scf, 'sd', costs):
            break
        for ik in range(atoms.kpts.Nk):
            for spin in range(atoms.occ.Nspin):
                g = grad(scf, ik, spin, W, **scf._precomputed)
                W[ik][spin] = W[ik][spin] - betat * g
    return costs, W


@name('preconditioned line minimization')
def pclm(scf, W, Nit, cost=scf_step_occ, grad=get_grad_occ, condition=check_convergence, betat=1,
         precondition=True, **kwargs):
    """Preconditioned line minimization algorithm for a fixed Hamiltonian.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        precondition (bool): Whether to use a preconditioner.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[list, ndarray]: Band energies per SCF cycle and optimized expansion coefficients.
    """
    atoms = scf.atoms
    costs = []

    linmin = np.empty((atoms.kpts.Nk, atoms.occ.Nspin))
    d = [np.empty_like(Wk) for Wk in W]

    if precondition:
        method = 'pclm'
    else:
        method = 'lm'

    for i in range(Nit):
        c = cost(scf, W)
        costs.append(c)
        if condition(scf, method, costs, linmin):
            break
        for ik in range(atoms.kpts.Nk):
            for spin in range(atoms.occ.Nspin):
                g = grad(scf, ik, spin, W, **scf._precomputed)
                if scf.log.level <= logging.DEBUG and i > 0:
                    linmin[ik][spin] = linmin_test(g, d[ik][spin])
                if precondition:
                    d[ik][spin] = -atoms.K(g, ik)
                else:
                    d[ik][spin] = -g
                W[ik][spin] = W[ik][spin] + betat * d[ik][spin]
                gt = grad(scf, ik, spin, W, **scf._precomputed)
                beta = betat * dotprod(g, d[ik][spin]) / dotprod(g - gt, d[ik][spin])
                W[ik][spin] = W[ik][spin] + beta * d[ik][spin]
    return costs, W


@name('line minimization')
def lm(scf, W, Nit, cost=scf_step_occ, grad=get_grad_occ, condition=check_convergence, betat=1,
       **kwargs):
    """Line minimization algorithm for a fixed Hamiltonian.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[list, ndarray]: Band energies per SCF cycle and optimized expansion coefficients.
    """
    return pclm(scf, W, Nit, cost, grad, condition, betat, precondition=False)


@name('preconditioned conjugate-gradient minimization')
def pccg(scf, W, Nit, cost=scf_step_occ, grad=get_grad_occ, condition=check_convergence, betat=1,
         cgform=1, precondition=True):
    """Preconditioned conjugate-gradient minimization algorithm for a fixed Hamiltonian.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        cgform (int): Conjugate gradient form.
        precondition (bool): Whether to use a preconditioner.

    Returns:
        tuple[list, ndarray]: Band energies per SCF cycle and optimized expansion coefficients.
    """
    atoms = scf.atoms
    costs = []

    linmin = np.empty((atoms.kpts.Nk, atoms.occ.Nspin))
    cg = np.empty((atoms.kpts.Nk, atoms.occ.Nspin))
    norm_g = np.empty((atoms.kpts.Nk, atoms.occ.Nspin))
    d = [np.empty_like(Wk) for Wk in W]
    d_old = [np.empty_like(Wk) for Wk in W]
    g_old = [np.empty_like(Wk) for Wk in W]

    if precondition:
        method = 'pccg'
    else:
        method = 'cg'

    c = cost(scf, W)
    costs.append(c)
    condition(scf, method, costs)
    # Do the first step without the linmin and cg tests, and without the cg_method
    for ik in range(atoms.kpts.Nk):
        for spin in range(atoms.occ.Nspin):
            g = grad(scf, ik, spin, W, **scf._precomputed)
            if precondition:
                d[ik][spin] = -atoms.K(g, ik)
            else:
                d[ik][spin] = -g
            W[ik][spin] = W[ik][spin] + betat * d[ik][spin]
            gt = grad(scf, ik, spin, W, **scf._precomputed)
            beta = betat * dotprod(g, d[ik][spin]) / dotprod(g - gt, d[ik][spin])
            g_old[ik][spin], d_old[ik][spin] = g, d[ik][spin]
            W[ik][spin] = W[ik][spin] + beta * d[ik][spin]

    for _ in range(1, Nit):
        c = cost(scf, W)
        costs.append(c)
        if condition(scf, method, costs, linmin, cg, norm_g):
            break
        for ik in range(atoms.kpts.Nk):
            for spin in range(atoms.occ.Nspin):
                g = grad(scf, ik, spin, W, **scf._precomputed)
                # Calculate linmin and cg for each spin separately
                if scf.log.level <= logging.DEBUG:
                    linmin[ik][spin] = linmin_test(g, d[ik][spin])
                    cg[ik][spin] = cg_test(atoms, ik, g, g_old[ik][spin], precondition)
                beta, norm_g[ik][spin] = cg_method(scf, ik, cgform, g, g_old[ik][spin],
                                                   d_old[ik][spin], precondition)
                if precondition:
                    d[ik][spin] = -atoms.K(g, ik) + beta * d_old[ik][spin]
                else:
                    d[ik][spin] = -g + beta * d_old[ik][spin]
                W[ik][spin] = W[ik][spin] + betat * d[ik][spin]
                gt = grad(scf, ik, spin, W, **scf._precomputed)
                beta = betat * dotprod(g, d[ik][spin]) / dotprod(g - gt, d[ik][spin])
                g_old[ik][spin], d_old[ik][spin] = g, d[ik][spin]
                W[ik][spin] = W[ik][spin] + beta * d[ik][spin]
    return costs, W


@name('conjugate-gradient minimization')
def cg(scf, W, Nit, cost=scf_step_occ, grad=get_grad_occ, condition=check_convergence, betat=1,
       cgform=1):
    """Conjugate-gradient minimization algorithm for a fixed Hamiltonian.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        cgform (int): Conjugate gradient form.

    Returns:
        tuple[list, ndarray]: Band energies per SCF cycle and optimized expansion coefficients.
    """
    return pccg(scf, W, Nit, cost, grad, condition, betat, cgform, precondition=False)


@name('auto minimization')
def auto(scf, W, Nit, cost=scf_step_occ, grad=get_grad_occ, condition=check_convergence, betat=1,
         cgform=1):
    """Automatic precond. conjugate-gradient minimization algorithm for a fixed Hamiltonian.

    This function chooses an sd step over the pccg step if the energy goes up.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): Step size.
        cgform (int): Conjugate gradient form.

    Returns:
        tuple[list, ndarray]: Band energies per SCF cycle and optimized expansion coefficients.
    """
    atoms = scf.atoms
    costs = []

    linmin = np.empty((atoms.kpts.Nk, atoms.occ.Nspin))
    cg = np.empty((atoms.kpts.Nk, atoms.occ.Nspin))
    norm_g = np.empty((atoms.kpts.Nk, atoms.occ.Nspin))
    g = [np.empty_like(Wk) for Wk in W]
    d = [np.empty_like(Wk) for Wk in W]
    d_old = [np.empty_like(Wk) for Wk in W]
    g_old = [np.empty_like(Wk) for Wk in W]

    # Do the first step without the linmin and cg tests, and without the cg_method
    for ik in range(atoms.kpts.Nk):
        for spin in range(atoms.occ.Nspin):
            g[ik][spin] = grad(scf, ik, spin, W, **scf._precomputed)
            d[ik][spin] = -atoms.K(g[ik][spin], ik)
            W[ik][spin] = W[ik][spin] + betat * d[ik][spin]
            gt = grad(scf, ik, spin, W, **scf._precomputed)
            beta = betat * dotprod(g[ik][spin], d[ik][spin]) / dotprod(g[ik][spin] - gt,
                                                                       d[ik][spin])
            g_old[ik][spin], d_old[ik][spin] = g[ik][spin], d[ik][spin]
            W[ik][spin] = W[ik][spin] + beta * d[ik][spin]

    c = cost(scf, W)
    costs.append(c)
    if condition(scf, 'pccg', costs):
        return costs

    for _ in range(1, Nit):
        W_old = copy.deepcopy(W)
        for ik in range(atoms.kpts.Nk):
            for spin in range(atoms.occ.Nspin):
                g[ik][spin] = grad(scf, ik, spin, W, **scf._precomputed)
                # Calculate linmin and cg for each spin separately
                if scf.log.level <= logging.DEBUG:
                    linmin[ik][spin] = linmin_test(g[ik][spin], d[ik][spin])
                    cg[ik][spin] = cg_test(atoms, ik, g[ik][spin], g_old[ik][spin])
                beta, norm_g[ik][spin] = cg_method(scf, ik, cgform, g[ik][spin], g_old[ik][spin],
                                                   d_old[ik][spin])
                d[ik][spin] = -atoms.K(g[ik][spin], ik) + beta * d_old[ik][spin]
                W[ik][spin] = W[ik][spin] + betat * d[ik][spin]
                gt = grad(scf, ik, spin, W, **scf._precomputed)
                beta = betat * dotprod(g[ik][spin], d[ik][spin]) / dotprod(g[ik][spin] - gt,
                                                                           d[ik][spin])
                g_old[ik][spin], d_old[ik][spin] = g[ik][spin], d[ik][spin]
                W[ik][spin] = W[ik][spin] + beta * d[ik][spin]

        c = cost(scf, W)
        # If the energy does not go down use the steepest descent step and recalculate the energy
        if c > costs[-1]:
            W = W_old
            for ik in range(atoms.kpts.Nk):
                for spin in range(atoms.occ.Nspin):
                    W[ik][spin] = W[ik][spin] - betat * g[ik][spin]
            c = cost(scf, W)
            costs.append(c)
            # Do not print cg and linmin if we do the sd step
            if condition(scf, 'sd', costs, norm_g=norm_g):
                break
        else:
            costs.append(c)
            if condition(scf, 'pccg', costs, linmin, cg, norm_g):
                break
    return costs, W


#: Map minimizer names with their respective implementation.
IMPLEMENTED = {
    'sd': sd,
    'lm': lm,
    'pclm': pclm,
    'cg': cg,
    'pccg': pccg,
    'auto': auto
}
